/* mkdep

   The present program constructs a list of makefile dependencies, looking
   at 'include' statements within FORTRAN programs.

   The program reads from standard input the list of files to be analyzed
   (with suffix .f or .h), searches within the files for statements like

		include "filename"

   and produces (on standard output) the relevant dependencies to be included
   in a makefile. Typical use of the program could be:

		ls *.f *.h | mkdep >> makefile

   A file.f containing include "file1", include "file2" will appear on output
   as
		file.o: file1 file2

   A file.h containing include "file1", include "file2" will appear on output
   as
		file.h: file1 file2
			touch file.h


   Options:  -v  (verbose): reports warnings on the use on non-standard
                            constructs in the FORTRAN files (very primitive)

   Return code:  0  -  OK
                 1  -  Problem in file opening; the execution is OK for the
                       other files
                 2  -  Illegal options
                 3  -  Illegal file name
                 4  -  Irregular syntax of include statements               */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#define N 200
#define LISTSIZE 500

/* Function declarations */

int iscomment();
int iscont();
int strdsp();
void updlist();
void additem();

/* External variables */

int nfile=0;
int nline;
int verbose=0;
char filename[N];
char recwsp[N];
char *progname;
char *quotes="\"\'";
char listdep[LISTSIZE];

main(argnum,args)
  int argnum;
  char *args[];
  { int retcode=0;
    int flanum;
    char target[N];

    /* Program name computation */

    if( (progname = strrchr(args[0],'/')) == (char *)0 ) progname = args[0];
    else progname++;

    /* Option handling */

    for (flanum=1; flanum < argnum; flanum++ )
      { int opt;
        if (*args[flanum] != '-')
         { (void)fprintf(stderr,
             "%s: %s: illegal option (not beginning with \"-\")\n",progname,
                     args[flanum]); return (2); }
        for (opt=1; opt < strlen(args[flanum]); opt++)
         { switch (*(args[flanum]+opt))
             { case 'v' : verbose = 1; break;
               default: (void)fprintf(stderr,"%s: %s: unknown option\n",
                        progname,args[flanum]); return (2); }  }  }

    /* Reads a file name from standard input: the separator is a "space",
       namely a blank, a TAB, a NEWLINE etc.  */

    while ( fscanf(stdin,"%s",filename) != EOF )
      { FILE *file;
        char rec[N];
        char *pdot;
        nfile++;

        /* Check on the filename extension and definition of the target */

        (void)strcpy(target,filename);
        if ( ((pdot=strchr(target,'.')) == (char *)0) || strlen(pdot) != 2 )
          { (void)fprintf(stderr,"%s: illegal file name %s\n",
                        progname,target);
            retcode=3;
            continue; }
        switch (*(++pdot))
           { case 'h': break;
             case 'f': *pdot='o'; break;
             default: (void)fprintf(stderr,"%s: illegal file name %s\n",
                        progname,target); retcode=3; continue; }

        /* File opening */

        nline=0;
        listdep[0]='\0';
        if( (file=fopen(filename,"r")) == NULL )
           { (void)fprintf(stderr,"%s: cannot open file %s\n",
                          progname,filename);
             retcode=1;
             continue; }

        /* Reads a record from the file; the separator is a newline character
          (/n), which is included in the string (rec) */

        while ( fgets(rec,N,file) != (char *)0 )
          { char *incl="include";
            char *item;
            char *incfile;
            nline++;
            if(iscomment(rec)) continue;
            if(iscont(rec)) continue;
            if(strdsp(rec,recwsp)<8) continue;
            item=strtok(recwsp,quotes);
            if (strcasecmp(item,incl) != 0) continue;
            if( (incfile=strtok((char *)0,quotes)) == (char *)0 )
               { (void)fprintf(stderr,
                  "%s: missing include file name, file %s, line %d\n",
                    progname,filename,nline); retcode=4; continue; }
            updlist(incfile); }

        nline=0;

        if ( listdep[0] != '\0' )
          { (void)fprintf(stdout,"%s:%s\n",target,listdep); 
             if ( *pdot=='h' ) (void)fprintf(stdout,"\ttouch %s\n\n",target);
             else (void)fprintf(stdout,"\n"); }

        (void)fclose(file);  }
    return(retcode); }

/* Check for a comment line */

int iscomment(s)
  char *s;
  { if ( (*s == 'c' ) || (*s == 'C' ) || (*s == '*' ) ) return 1;
    else if ( !isspace (*s) && verbose )
          (void)fprintf(stderr,
                  "%s: first character is not a blank, file %s, line %d\n",
                    progname,filename,nline);
    return 0; }

/* Check for a continuation line */

int iscont(s)
  char *s;
  { int c;
    for (c=0;c<5;c++)
      { if ( *(s+c) == '\t' ) return 0; }
    if ( *(s+5) != ' ' ) return 1;
    return 0; }

/* Delete space characters from string s, giving swsp; returns the length of
   the reduced string */

int strdsp(s,swsp)
  char *s;
  char *swsp;
  { int c;
    int c1=0;
    for (c=0;c<strlen(s);c++)
      { if ( !isspace(*(s+c)) ) *(swsp+(c1++)) =  *(s+c) ; }
    *(swsp+c1) = '\0' ; return c1; }

/* Add a new dependency to the list of dependencies */

void updlist(dep)
  char *dep;	/* Dependency to be added to the list */
   { char *blank=" ";
     char scan[LISTSIZE];
     char *scanp;
     char *entry;
     scanp=strcpy(scan,listdep);
     for ( entry=strtok(scanp,blank); entry!=(char *)0;
				entry=strtok((char *)0,blank))
       { if (strcmp(entry,dep) == 0) return; }
     additem(dep); }

/* Add a new item to the list listdep */

void additem(newitem)
  char *newitem;
   { int i;
     int k;
     int ll=strlen(listdep);
     int ln=strlen(newitem);
     int sizreq=ll+ln+2;
     assert(sizreq<=LISTSIZE);
     listdep[ll]=' ';
     for (i=ll+1,k=0;k<ln;i++,k++) listdep[i] = newitem[k];
     listdep[i]='\0';
     return; }

